"""Exceptions module."""


class HomerError(Exception):
    """Parent exception class for all Homer exceptions."""


class HomerAbortError(HomerError):
    """Exception class for aborted actions."""


class HomerTimeoutError(HomerError):
    """Exception class for actions that timeout."""



"""Homer package."""
import logging
import os
import pathlib
import sys

from collections import defaultdict
from importlib import import_module
from typing import Callable, DefaultDict, Dict, List, Mapping, Optional, Tuple

import pynetbox

from pkg_resources import DistributionNotFound, get_distribution

from homer.capirca import CapircaGenerate
from homer.config import HierarchicalConfig, load_yaml_config
from homer.devices import Device, Devices
from homer.exceptions import HomerAbortError, HomerError, HomerTimeoutError
from homer.netbox import NetboxData, NetboxDeviceData, NetboxInventory
from homer.templates import Renderer
from homer.transports.junos import connected_device


TIMEOUT_ATTEMPTS = 3
""":py:class:`int`: the number of attempts to try when there is a timeout."""
DIFF_EXIT_CODE = 99
""":py:class:`int`: the exit code used when the diff command is executed and there is a diff."""


try:
    __version__ = get_distribution('homer').version  # Must be the same used as 'name' in setup.py
    """:py:class:`str`: the version of the current Homer package."""
except DistributionNotFound:  # pragma: no cover - this should never happen during tests
    pass  # package is not installed

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Homer:
    """The instance to run Homer."""

    OUT_EXTENSION = '.out'
    """:py:class:`str`: the extension for the generated output files."""

    def __init__(self, main_config: Mapping):
        """Initialize the instance.

        Arguments:
            main_config (dict): the configuration dictionary.

        """
        logger.debug('Initialized with configuration: %s', main_config)
        self._main_config = main_config
        self.private_base_path = self._main_config['base_paths'].get('private', '')
        self._config = HierarchicalConfig(
            self._main_config['base_paths']['public'], private_base_path=self.private_base_path)

        self._netbox_api = None
        self._device_plugin = None
        if self._main_config.get('netbox', {}):
            self._netbox_api = pynetbox.api(
                self._main_config['netbox']['url'], token=self._main_config['netbox']['token'])
            if self._main_config['netbox'].get('plugin', ''):
                self._device_plugin = import_module(  # type: ignore
                    self._main_config['netbox']['plugin']).NetboxDeviceDataPlugin

        devices_all_config = load_yaml_config(
            os.path.join(self._main_config['base_paths']['public'], 'config', 'devices.yaml'))
        devices_config = {fqdn: data.get('config', {}) for fqdn, data in devices_all_config.items()}

        netbox_inventory = self._main_config.get('netbox', {}).get('inventory', {})
        if netbox_inventory:
            devices = NetboxInventory(
                self._netbox_api, netbox_inventory['device_roles'], netbox_inventory['device_statuses']).get_devices()
        else:
            devices = devices_all_config.copy()
            for data in devices.values():
                data.pop('config', None)

        private_devices_config: Dict = {}
        if self.private_base_path:
            private_devices_config = load_yaml_config(
                os.path.join(self.private_base_path, 'config', 'devices.yaml'))

        self._ignore_warning = self._main_config.get('transports', {}).get('junos', {}).get('ignore_warning', False)
        self._transport_username = self._main_config.get('transports', {}).get('username', '')
        self._transport_ssh_config = self._main_config.get('transports', {}).get('ssh_config', None)
        self._devices = Devices(devices, devices_config, private_devices_config)
        self._renderer = Renderer(self._main_config['base_paths']['public'], self.private_base_path)
        self._output_base_path = pathlib.Path(self._main_config['base_paths']['output'])

    def generate(self, query: str) -> int:
        """Generate the configuration only saving it locally, no remote action is performed.

        Arguments:
            query (str): the query to select the devices.

        Return:
            int: ``0`` on success, a small positive integer on failure.

        """
        logger.info('Generating configuration for query %s', query)
        self._prepare_out_dir()
        successes, _ = self._execute(self._device_generate, query)
        return Homer._parse_results(successes)

    def diff(self, query: str, *, omit_diff: bool = False) -> int:
        """Generate the configuration and check the diff with the current live one.

        Arguments:
            query (str): the query to select the devices.
            omit_diff (bool, optional): whether to not show the actual diff to avoid leak of private data.

        Return:
            int: ``0`` on success, a small positive integer on failure.

        """
        logger.info('Generating diff for query %s', query)
        successes, diffs = self._execute(self._device_diff, query)
        has_diff = False
        for diff, diff_devices in diffs.items():
            print('Changes for {n} devices: {devices}'.format(n=len(diff_devices), devices=diff_devices))
            if diff is None:
                print('# Failed')
            elif not diff:
                print('# No diff')
            else:
                has_diff = True
                if omit_diff:
                    print('# Non-empty diff omitted, -o/--omit-diff set')
                else:
                    print(diff)
            print('---------------')

        ret = Homer._parse_results(successes)
        if ret == 0 and has_diff:
            return DIFF_EXIT_CODE

        return ret

    def commit(self, query: str, *, message: str = '-') -> int:
        """Commit the generated configuration asking for confirmation.

        Arguments:
            query (str): the query to select the devices.
            message (str): the commit message to use.

        Return:
            int: ``0`` on success, a small positive integer on failure.

        """
        logger.info('Committing config for query %s with message: %s', query, message)
        successes, _ = self._execute(self._device_commit, query, message=message)
        return Homer._parse_results(successes)

    def _device_generate(self, device: Device, device_config: str, _: int) -> Tuple[bool, Optional[str]]:
        """Save the generated configuration in a local file.

        Arguments:
            device (homer.devices.Device): the device instance.
            device_config (str): the generated configuration for the device.
            attempt (int, unused): the current attempt number.

        Returns:
            tuple: a two-element tuple with a boolean as first parameter that represent the success of the operation
            or not and a second element with a string or None that is not used but is required by the callback API.

        """
        output_path = self._output_base_path / '{fqdn}{out}'.format(fqdn=device.fqdn, out=Homer.OUT_EXTENSION)
        with open(str(output_path), 'w') as f:
            f.write(device_config)
            logger.info('Written configuration for %s in %s', device.fqdn, output_path)

        return True, None

    def _device_diff(self, device: Device, device_config: str,
                     _: int) -> Tuple[bool, Optional[str]]:  # pylint: disable=no-self-use
        """Perform a configuration diff between the generated configuration and the live one.

        Arguments:
            device (homer.devices.Device): the device instance.
            device_config (str): the generated configuration for the device.
            attempt (int, unused): the current attempt number.

        Returns:
            tuple: a two-element tuple with a boolean as first parameter that represent the success of the operation
            or not and a second element with a string that contains the configuration differences or None if unable
            to load the new configuration in the device to generate the diff.

        """
        with connected_device(device.fqdn, username=self._transport_username,
                              ssh_config=self._transport_ssh_config) as connection:
            return connection.commit_check(device_config, self._ignore_warning)

    def _device_commit(self, device: Device, device_config: str,  # noqa: MC0001; pylint: disable=no-self-use
                       attempt: int, *, message: str = '-') -> Tuple[bool, Optional[str]]:
        """Commit a new configuration to the device.

        Arguments:
            device (homer.devices.Device): the device instance.
            device_config (str): the generated configuration for the device.
            attempt (int): the current attempt number.
            message (str): the commit message to use.

        Raises:
            HomerTimeoutError: on timeout.

        Returns:
            tuple: a two-element tuple with a boolean as first parameter that represent the success of the operation
            or not and a second element with a string or None that is not used but is required by the callback API.

        """
        def callback(fqdn: str, diff: str) -> None:
            """Callback as required by :py:class:`homer.transports.junos.ConnectedDevice.commit`."""
            if not sys.stdout.isatty():
                raise HomerError('Not in a TTY, unable to ask for confirmation')

            print('Configuration diff for {fqdn}:\n{diff}'.format(fqdn=fqdn, diff=diff))
            print('Type "yes" to commit, "no" to abort.')

            for _ in range(2):
                resp = input('> ')
                if resp == 'yes':
                    break
                if resp == 'no':
                    raise HomerAbortError('Commit aborted')

                print(('Invalid response, please type "yes" to commit or "no" to abort. After 2 wrong answers the '
                       'commit will be aborted.'))
            else:
                raise HomerAbortError('Too many invalid answers, commit aborted')

        is_retry = (attempt != 1)
        with connected_device(device.fqdn, username=self._transport_username,
                              ssh_config=self._transport_ssh_config) as connection:
            try:
                connection.commit(device_config, message, callback, ignore_warning=self._ignore_warning,
                                  is_retry=is_retry)
                return True, ''
            except HomerTimeoutError:
                raise  # To be catched later for automatic retry
            except HomerAbortError as e:
                logger.warning('%s on %s', e, device.fqdn)
            except Exception as e:  # pylint: disable=broad-except
                logger.error('Failed to commit on %s: %s', device.fqdn, e)
                logger.debug('Traceback:', exc_info=True)

            return False, ''

    def _prepare_out_dir(self) -> None:
        """Prepare the out directory creating the directory if doesn't exists and deleting any pre-generated file."""
        self._output_base_path.mkdir(parents=True, exist_ok=True)
        for path in self._output_base_path.iterdir():
            if path.is_file() and path.suffix == Homer.OUT_EXTENSION:
                path.unlink()

    def _execute(self, callback: Callable, query: str, **kwargs: str) -> Tuple[Dict, DefaultDict]:  # noqa, mccabe: MC0001 too complex
        """Execute Homer based on the given action and query.

        Arguments:
            callback (Callable): the callback to call for each device.
            query (str): the query to filter the devices to act on.
            **kwargs (str): any additional keyword argument to pass to the callback

        Returns:
            tuple: a two-element tuple, with the first item as a dictionary that contains two keys (:py:data:`True`
            and :py:data:`False`) and as value a list of device FQDN that were successful (True) or failed (False)
            the operation and a second element a :py:class:`collections.defaultdict` that has as keys the
            configuration differences and as values the list of device FQDN that reported that diff.

        """
        diffs: DefaultDict[str, list] = defaultdict(list)
        successes: Dict[bool, list] = {True: [], False: []}
        netbox_data = None
        if self._netbox_api is not None:
            logger.info('Gathering global Netbox data')
            netbox_data = NetboxData(self._netbox_api)

        for device in self._devices.query(query):
            logger.info('Generating configuration for %s', device.fqdn)

            try:
                device_config = []
                device_data = self._config.get(device)
                # Render the ACLs using Capirca
                if 'capirca' in device_data:
                    capirca = CapircaGenerate(self._main_config, device_data['capirca'], self._netbox_api)
                    generated_acls = capirca.generate_acls()
                    if generated_acls:
                        device_config.extend(generated_acls)

                if netbox_data is not None:
                    device_data['netbox'] = {
                        'global': netbox_data,
                        'device': NetboxDeviceData(self._netbox_api, device),
                    }
                    if self._device_plugin is not None:
                        device_data['netbox']['device_plugin'] = self._device_plugin(self._netbox_api, device)
                # Render the Jinja templates based on yaml + netbox data
                device_config.append(self._renderer.render(device.metadata['role'], device_data))
            except HomerError:
                logger.exception('Device %s failed to render the template, skipping.', device.fqdn)
                successes[False].append(device.fqdn)
                continue

            for attempt in range(1, TIMEOUT_ATTEMPTS + 1):
                try:
                    device_success, device_diff = callback(device, '\n'.join(device_config), attempt, **kwargs)
                    break
                except HomerTimeoutError as e:
                    logger.error('Commit attempt %d/%d failed: %s', attempt, TIMEOUT_ATTEMPTS, e)
            else:
                device_success = False
                device_diff = ''

            successes[device_success].append(device.fqdn)
            diffs[device_diff].append(device.fqdn)

        return successes, diffs

    @staticmethod
    def _parse_results(successes: Mapping[bool, List[Device]]) -> int:
        """Parse the results dictionary, log and return the approriate exit status code.

        Arguments:
            successes (dict): a dictionary that contains two keys (:py:data:`True` and :py:data:`False`) and as value
                a list of device FQDN that were successful (True) or failed (False) the operation.

        Return:
            int: ``0`` on success, a small positive integer on failure.

        """
        if successes[False]:
            logger.error('Homer run had issues on %d devices: %s', len(successes[False]), successes[False])
            return 1

        logger.info('Homer run completed successfully on %d devices: %s', len(successes[True]), successes[True])
        return 0



"""Netbox module tests."""
# pylint: disable=attribute-defined-outside-init
from collections import UserDict
from unittest import mock

import pytest

from homer.devices import Device
from homer.exceptions import HomerError
from homer.netbox import BaseNetboxData, NetboxData, NetboxDeviceData, NetboxInventory


class NetboxObject:  # pylint: disable=too-many-instance-attributes
    """Helper class to mimic pynetbox objects."""

    def __iter__(self):
        """Allow to convert the object to dict."""
        return iter(vars(self).items())


def mock_netbox_device(name, role, site, status,  # pylint: disable=too-many-arguments,too-many-branches
                       ip4=False, ip6=False, virtual_chassis=False, platform=True):
    """Returns a mocked Netbox device object."""
    device = NetboxObject()
    device.id = 123  # pylint: disable=invalid-name
    device.name = name
    device.device_role = NetboxObject()
    device.device_role.slug = role
    device.site = NetboxObject()
    device.site.slug = site
    device.status = NetboxObject()
    device.status.value = status.lower()
    device.status.label = status
    device.device_type = NetboxObject()
    device.device_type.slug = 'typeA'

    if ip4:
        device.primary_ip4 = NetboxObject()
        device.primary_ip4.address = '127.0.0.1/32'
        device.primary_ip4.dns_name = name
    else:
        device.primary_ip4 = None

    if ip6:
        device.primary_ip6 = NetboxObject()
        device.primary_ip6.address = '::1/128'
        device.primary_ip6.dns_name = name
    else:
        device.primary_ip6 = None

    if virtual_chassis:
        device.virtual_chassis = NetboxObject()
        device.virtual_chassis.id = 1  # pylint: disable=invalid-name
    else:
        device.virtual_chassis = None

    if platform:
        device.platform = NetboxObject()
        device.platform.slug = 'osA'
    else:
        device.platform = None

    return device


def mock_netbox_virtual_chassis(master, domain):
    """Returns a mocked Netbox virtual chassis object."""
    vc = NetboxObject()
    vc.master = master
    vc.domain = domain
    return vc


class TestBaseNetboxData:
    """BaseNetboxData class tests."""

    def setup_method(self):
        """Initialize the test instances."""
        self.netbox_api = mock.MagicMock()  # Can't use spec_set because of pynetbox lazy creation
        self.netbox_data = BaseNetboxData(self.netbox_api)

        def key_raise():
            raise RuntimeError('key raise')

        self.netbox_data._get_key_raise = key_raise  # pylint: disable=protected-access

    def test_init(self):
        """An instance of BaseNetboxData should be also an instance of UserDict."""
        assert isinstance(self.netbox_data, BaseNetboxData)
        assert isinstance(self.netbox_data, UserDict)

    def test_getitem_fail(self):
        """Should raise KeyError if there is no method for that key."""
        with pytest.raises(KeyError, match='key1'):
            self.netbox_data['key1']  # pylint: disable=pointless-statement

    def test_getitem_raise(self):
        """Should raise HomerError if the call to the key raise any exception."""
        with pytest.raises(HomerError, match='Failed to get key key_raise'):
            self.netbox_data['key_raise']  # pylint: disable=pointless-statement


class TestNetboxData:
    """NetboxData class tests."""

    def setup_method(self):
        """Initialize the test instances."""
        self.netbox_api = mock.MagicMock()  # Can't use spec_set because of pynetbox lazy creation
        self.netbox_data = NetboxData(self.netbox_api)

    def test_init(self):
        """An instance of NetboxData should be also an instance of UserDict."""
        assert isinstance(self.netbox_data, NetboxData)
        assert isinstance(self.netbox_data, UserDict)

    def test_vlans(self):
        """It should return the defined vlans."""
        vlan = NetboxObject()
        vlan.id = 1
        self.netbox_api.ipam.vlans.all.return_value = [vlan]
        assert self.netbox_data['vlans'] == [{'id': 1}]


class TestNetboxDeviceData:
    """NetboxDeviceData class tests."""

    def setup_method(self):
        """Initialize the test instances."""
        self.netbox_api = mock.MagicMock()  # Can't use spec_set because of pynetbox lazy creation
        netbox_device = mock_netbox_device('device1.example.com', 'role1', 'site1', 'Active')
        self.device = Device(netbox_device.name, {'netbox_object': netbox_device}, {}, {})
        self.netbox_data = NetboxDeviceData(self.netbox_api, self.device)

    def test_init(self):
        """An instance of NetboxData should be also an instance of UserDict."""
        assert isinstance(self.netbox_data, NetboxDeviceData)
        assert isinstance(self.netbox_data, UserDict)

    def test_cached_key(self):
        """If a key has been already populated it should not call the method again."""

    def test_get_virtual_chassis_members_no_members(self):
        """If a device is not part of a virtual chassis it should return None."""
        assert self.netbox_data['virtual_chassis_members'] is None

    def test_get_virtual_chassis_members_with_members(self):
        """If a device is part of a virtual chassis it should return its members."""
        netbox_devices = []
        devices = []
        names = ('device1.example.com', 'device2.example.com')
        for name in names:
            netbox_device = mock_netbox_device(name, 'role1', 'site1', 'Active', virtual_chassis=True)
            netbox_devices.append(netbox_device)
            devices.append(Device(netbox_device.name, {'netbox_object': netbox_device}, {}, {}))

        self.netbox_api.dcim.devices.filter.return_value = netbox_devices

        netbox_data = NetboxDeviceData(self.netbox_api, devices[0])
        # Call multiple times so that we can check the results are cached
        for _ in range(2):
            assert [d['name'] for d in netbox_data['virtual_chassis_members']] == list(names)

        # Ensure the data was cached and the API was called only once.
        assert self.netbox_api.dcim.devices.filter.call_count == 1

    def test_get_inventory(self):
        """It should return all the inventory items of a device."""
        netbox_object = NetboxObject()
        netbox_object.item = 'inventory item'
        self.netbox_api.dcim.inventory_items.filter.return_value = [netbox_object]

        assert self.netbox_data['inventory'] == [{'item': netbox_object.item}]
        self.netbox_api.dcim.inventory_items.filter.assert_called_once_with(device_id=123)

    def test_get_circuits(self):
        """It should return all the inventory items of a device."""
        cable_1 = NetboxObject()
        cable_1.termination_a_type = 'circuits.circuittermination'
        cable_1.termination_b = NetboxObject()
        cable_1.termination_b.name = 'int1'
        cable_1.termination_a = NetboxObject()
        cable_1.termination_a.circuit = NetboxObject()
        cable_1.termination_a.circuit.id = 1
        cable_2 = NetboxObject()
        cable_2.termination_a_type = 'unused'
        cable_2.termination_b_type = 'circuits.circuittermination'
        cable_2.termination_a = NetboxObject()
        cable_2.termination_a.name = 'int2'
        cable_2.termination_b = NetboxObject()
        cable_2.termination_b.circuit = NetboxObject()
        cable_2.termination_b.circuit.id = 2
        cable_3 = NetboxObject()
        cable_3.termination_a_type = 'unused'
        cable_3.termination_b_type = 'unused'

        self.netbox_api.circuits.circuits.get.return_value = {}
        self.netbox_api.dcim.cables.filter.return_value = [cable_1, cable_2, cable_3]

        assert self.netbox_data['circuits'] == {'int1': {}, 'int2': {}}
        self.netbox_api.dcim.cables.filter.assert_called_once_with(device_id=123)
        calls = [mock.call(1), mock.call(2)]
        self.netbox_api.circuits.circuits.get.assert_has_calls(calls)

    def test_get_vlans(self):
        """It should return the vlans defined on a device's interfaces."""
        interface1 = NetboxObject()  # This is a fake interface object
        interface1.untagged_vlan = NetboxObject()  # That interface have fake tagged and untagged vlans
        interface1.untagged_vlan.vid = 666
        interface1.tagged_vlans = [NetboxObject(), NetboxObject()]
        interface1.tagged_vlans[0].vid = 667
        interface1.tagged_vlans[1].vid = 667
        interface2 = NetboxObject()  # This is a fake interface object
        interface2.untagged_vlan = NetboxObject()  # That interface have fake tagged and untagged vlans
        interface2.untagged_vlan.vid = 666
        interface2.tagged_vlans = None
        self.netbox_api.dcim.interfaces.filter.return_value = [interface1, interface2]
        # We want the function with the fake inbound data to match the one untagged and tagged vlans
        assert self.netbox_data['vlans'] == {666: interface1.untagged_vlan, 667: interface1.tagged_vlans[0]}

        # And we want the fake API to be called only once
        self.netbox_api.dcim.interfaces.filter.assert_called_once_with(device_id=123)


class TestNetboxInventory:
    """NetboxInventory class tests."""

    def setup_method(self):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        selected_devices = [
            mock_netbox_device('device1', 'roleA', 'siteA', 'Active', ip4=True),
            mock_netbox_device('device2', 'roleA', 'siteA', 'Staged', ip6=True),
        ]
        selected_vcs = [
            mock_netbox_device('device1-vc1', 'roleA', 'siteA', 'Active'),
        ]
        filtered_devices = [
            mock_netbox_device('device3', 'roleB', 'siteB', 'Active'),
            mock_netbox_device('device4', 'roleA', 'siteA', 'Active'),
            mock_netbox_device('device2', 'roleA', 'siteA', 'Staged', platform=False),
        ]
        filtered_vcs = [
            mock_netbox_device('device1-vc2', 'roleB', 'siteA', 'Active'),
            mock_netbox_device('device1-vc3', 'roleA', 'siteA', 'Decommissioning'),
            mock_netbox_device('', 'roleA', 'siteA', 'Active'),
        ]
        self.selected_devices = selected_devices + selected_vcs
        virtual_chassis = [mock_netbox_virtual_chassis(device, device.name) for device in selected_vcs + filtered_vcs]
        self.mocked_api = mock.MagicMock()
        self.mocked_api.dcim.virtual_chassis.all.return_value = virtual_chassis
        self.mocked_api.dcim.devices.filter.return_value = selected_devices + filtered_devices
        self.inventory = NetboxInventory(self.mocked_api, ['roleA'], ['Active', 'Staged'])

    def test_get_devices(self):
        """It should get the devices without inspecting virtual chassis."""
        devices = self.inventory.get_devices()
        expected = {}
        for device in self.selected_devices:
            expected_device = {'site': device.site.slug, 'role': device.device_role.slug, 'type': 'typeA',
                               'netbox_object': device}
            if device.primary_ip4 is not None:
                expected_device['ip4'] = '127.0.0.1'
            if device.primary_ip6 is not None:
                expected_device['ip6'] = '::1'

            expected[device.name] = expected_device

        assert devices == expected



"""Devices module tests."""
from collections import UserDict

from homer.config import load_yaml_config
from homer.devices import Device, Devices
from homer.tests import get_fixture_path


class TestDevices:
    """Devices class tests."""

    def setup_method(self):
        """Initialize the test instances."""
        # pylint: disable=attribute-defined-outside-init
        devices = load_yaml_config(get_fixture_path('public', 'config', 'devices.yaml'))
        devices_config = {fqdn: device.get('config', {}) for fqdn, device in devices.items()}
        self.devices = Devices(devices, devices_config)
        self.devices_with_private = Devices(
            devices, devices_config, load_yaml_config(get_fixture_path('private', 'config', 'devices.yaml')))

    def test_init(self):
        """An instance of Devices should be also an instance of UserDict."""
        assert isinstance(self.devices, Devices)
        assert isinstance(self.devices, UserDict)

    def test_dict_access(self):
        """Should return the device with the given FQDN."""
        device = self.devices['device1.example.com']
        assert isinstance(device, Device)

    def test_private_config(self):
        """Should include the device-specific private config only when set."""
        assert self.devices['device1.example.com'].private == {}
        private_value = self.devices_with_private['device1.example.com'].private['device_private_key']
        assert private_value == 'device1_private_value'

    def test_query_role(self):
        """Should return all the devices with a given role."""
        devices = self.devices.query('role:roleA')
        assert len(devices) == 2
        for device in devices:
            assert isinstance(device, Device)
        assert sorted([device.fqdn for device in devices]) == ['another.example.com', 'device1.example.com']

    def test_query_role_single(self):
        """Should return all the devices with a given role, testing one device match."""
        devices = self.devices.query('role:roleB')
        assert len(devices) == 1
        assert isinstance(devices[0], Device)
        assert devices[0].fqdn == 'device2.example.com'

    def test_query_role_no_match(self):
        """Should return an empty list if no device has that role."""
        devices = self.devices.query('role:non-existent')
        assert devices == []

    def test_query_fqdn(self):
        """Should return the matching device."""
        devices = self.devices.query('device1.example.com')
        assert len(devices) == 1
        assert isinstance(devices[0], Device)
        assert devices[0].fqdn == 'device1.example.com'

    def test_query_fqdn_no_match(self):
        """Should return an empty list if not match."""
        devices = self.devices.query('non-existent.example.com')
        assert devices == []

    def test_query_fqdn_globbing(self):
        """Should return the matching devices."""
        devices = self.devices.query('device*.*.com')
        assert len(devices) == 2
        assert sorted([device.fqdn for device in devices]) == ['device1.example.com', 'device2.example.com']



"""JunOS module."""
import logging

from contextlib import contextmanager
from typing import Callable, Iterator, List, Optional, Tuple, Union

from jnpr.junos import Device as JunOSDevice
from jnpr.junos.exception import CommitError, ConfigLoadError, RpcTimeoutError, UnlockError
from jnpr.junos.utils.config import Config
from ncclient.operations.errors import TimeoutExpiredError

from homer.exceptions import HomerError, HomerTimeoutError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
DIFF_ADDED_CODE = 32
DIFF_REMOVED_CODE = 31
DIFF_MOVED_CODE = 33


@contextmanager
def connected_device(fqdn: str, *, username: str = '', ssh_config: Optional[str] = None) -> Iterator['ConnectedDevice']:
    """Context manager to perform actions on a connected Juniper device.

    Arguments:
        fqdn (str): the FQDN of the Juniper device.
        username (str): the username to use to connect to the Juniper device.
        ssh_config (Optional[str]): an ssh_config file if you want other than ~/.ssh/config

    Yields:
        ConnectedDevice: the Juniper connected device instance.

    """
    device = ConnectedDevice(fqdn, username=username, ssh_config=ssh_config)
    try:
        yield device
    finally:
        device.close()


# pylint: disable=no-member
# Pylint doesn't recognize the 'cu' member in junos.Device instances and generated-members seems to have a bug
class ConnectedDevice:
    """Juniper transport to manage a JunOS connected device."""

    def __init__(self, fqdn: str, *, username: str = '', ssh_config: Optional[str] = None):
        """Initialize the instance and open the connection to the device.

        Arguments:
            fqdn (str): the FQDN of the Juniper device.
            username (str): the username to use to connect to the Juniper device.
            ssh_config (Optional[str]): an ssh_config file if you want other than ~/.ssh/config

        """
        self._fqdn = fqdn
        logger.debug('Connecting to device %s (user %s ssh_config %s)', self._fqdn, username, ssh_config)
        self._device = JunOSDevice(host=self._fqdn, user=username, port=22, ssh_config=ssh_config)
        self._device.open()
        self._device.bind(cu=Config)

    def commit(self, config: str, message: str, callback: Callable, *,  # noqa, mccabe: MC0001 too complex (11)
               ignore_warning: Union[bool, str, List[str]] = False, is_retry: bool = False) -> None:
        """Commit the loaded configuration.

        Arguments:
            config (str): the device new configuration.
            message (str): the commit message to use.
            callback (callable): a callable function that accepts two parameters: a string with the FQDN of the
                current device and a string with the diff between the current configuration and the new one. The
                callback must raise any exception if the execution should be interrupted and the config rollbacked or
                return :py:data:`None`.
            ignore_warning (mixed, optional): the warnings to tell JunOS to ignore, see:
                https://junos-pyez.readthedocs.io/en/2.3.0/jnpr.junos.utils.html#jnpr.junos.utils.config.Config.load
            is_retry (bool, optional): whether this is a retry and the commit_check should be run anyway, also if the
                diff is empty.

        Raises:
            HomerTimeoutError: on timeout.
            HomerError: on commit error.
            Exception: on generic failure.

        """
        diff = self._prepare(config, ignore_warning)
        if not diff:
            if not is_retry:
                logger.info('Empty diff for %s, skipping device.', self._fqdn)
                return
        else:
            try:
                callback(self._fqdn, diff)
            except Exception:
                self._rollback()
                raise

        logger.info('Committing the configuration on %s', self._fqdn)
        try:
            if diff:
                self._device.cu.commit(confirm=2, comment=message)
            self._device.cu.commit_check()
        except RpcTimeoutError as e:
            raise HomerTimeoutError(str(e))
        except CommitError as e:
            raise HomerError('Commit error: {err}'.format(err=ConnectedDevice._parse_commit_error(e))) from e

    def commit_check(self, config: str,
                     ignore_warning: Union[bool, str, List[str]] = False) -> Tuple[bool, Optional[str]]:
        """Perform commit check, reuturn the diff and rollback.

        Arguments:
            config (str): the device new configuration.

        Returns:
            tuple: a two-element tuple with a boolean as first item that is :py:data:`True` on success and
            :py:data:`False` on failure and a string as second item with the difference between the current
            configuration and the new one, empty string on no diff and :py:data:`None` on failure.

        """
        success = False
        try:
            diff = self._prepare(config, ignore_warning)
        except Exception as e:  # pylint: disable=broad-except
            logger.error('Failed to get diff for %s: %s', self._fqdn, e)
            logger.debug('Traceback:', exc_info=True)
            self._rollback()
            return False, None

        if not diff:
            logger.info('Empty diff for %s, skipping device.', self._fqdn)
            self._rollback()
            return True, diff

        logger.info('Running commit check on %s', self._fqdn)
        try:
            self._device.cu.commit_check()
            success = True
        except CommitError as e:
            logger.error('Commit check error on %s: %s', self._fqdn, ConnectedDevice._parse_commit_error(e))
        except Exception as e:  # pylint: disable=broad-except
            logger.error('Failed to commit check on %s: %s', self._fqdn, e)
            logger.debug('Traceback:', exc_info=True)
        finally:
            self._rollback()

        return success, diff

    def close(self) -> None:
        """Close the connection with the device."""
        try:
            self._device.cu.unlock()
        except UnlockError:
            pass
        try:
            self._device.close()
        except TimeoutExpiredError:
            logger.warning('Unable to close the connection to the device: TimeoutExpiredError')

    def _prepare(self, config: str, ignore_warning: Union[bool, str, List[str]] = False) -> str:
        """Prepare the new configuration to be committed.

        Arguments:
            config (str): the device new configuration.

        Raises:
            Exception: on generic failure.

        Returns:
            str: the differences between the current config and the new one.

        """
        logger.debug('Preparing the configuration on %s', self._fqdn)
        diff = ''
        try:
            self._device.cu.lock()
            self._device.cu.load(config, format='text', merge=False, ignore_warning=ignore_warning)
            diff = self._device.cu.diff()
        except ConfigLoadError:
            raise
        except Exception:
            self._rollback()
            raise

        if diff is None:
            diff = ''

        return color_diff(diff)

    def _rollback(self) -> None:
        """Rollback the current staged configuration."""
        logger.debug('Rolling back staged config on %s', self._fqdn)
        try:
            self._device.cu.rollback()
        except ValueError as e:
            logger.error('Invalid rollback ID on %s: %s', self._fqdn, e)
        except Exception as e:  # pylint: disable=broad-except
            logger.error('Failed to rollback on %s: %s', self._fqdn, e)
            logger.debug('Traceback:', exc_info=True)

    @staticmethod
    def _parse_commit_error(exc: CommitError) -> str:
        """Parse a CommitError exception and returnonly the reason.

        Arguments:
            exc (jnpr.junos.exception.CommitError): the exception to parse.

        Returns:
            str: the reason for the commit errror.

        """
        if exc.rsp.find('.//ok') is None:
            return exc.rsp.findtext('.//error-message')
        return str(exc)


def color_diff(diff: str) -> str:
    """Color the diff based on JunOS diff syntax."""
    lines = []
    for line in diff.splitlines():
        if line.startswith('+'):
            code = DIFF_ADDED_CODE
        elif line.startswith('-'):
            code = DIFF_REMOVED_CODE
        elif line.startswith('!'):
            code = DIFF_MOVED_CODE
        else:
            code = 0

        if code:
            line = '\x1b[{code}m{line}\x1b[39m'.format(code=code, line=line)

        lines.append(line)

    colored_diff = '\n'.join(lines)
    if diff and diff[-1] == '\n':  # Re-add the last trailing newline if present
        colored_diff += '\n'

    return colored_diff



"""Netbox module."""
import ipaddress
import logging

from collections import UserDict
from typing import Any, Dict, List, Optional, Sequence

import pynetbox

from homer.devices import Device
from homer.exceptions import HomerError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class BaseNetboxData(UserDict):  # pylint: disable=too-many-ancestors
    """Base class to gather data dynamically from Netbox."""

    def __init__(self, api: pynetbox.api):
        """Initialize the dictionary.

        Arguments:
            api (pynetbox.api): the Netbox API instance.

        """
        super().__init__()
        self._api = api

    def __getitem__(self, key: Any) -> Any:
        """Dynamically call the related method, if exists, to return the requested data.

        Parameters according to Python's datamodel, see:
        https://docs.python.org/3/reference/datamodel.html#object.__getitem__

        Returns:
            mixed: the dynamically gathered data.

        """
        method_name = '_get_{key}'.format(key=key)
        if not hasattr(self, method_name):
            raise KeyError(key)

        if key not in self.data:
            try:
                self.data[key] = getattr(self, method_name)()
            except Exception as e:
                raise HomerError('Failed to get key {key}'.format(key=key)) from e

        return self.data[key]


class BaseNetboxDeviceData(BaseNetboxData):  # pylint: disable=too-many-ancestors
    """Base class to gather device-specific data dynamically from Netbox."""

    def __init__(self, api: pynetbox.api, device: Device):
        """Initialize the dictionary.

        Arguments:
            api (pynetbox.api): the Netbox API instance.
            device (homer.devices.Device): the device for which to gather the data.

        """
        super().__init__(api)
        self._device = device


class NetboxData(BaseNetboxData):  # pylint: disable=too-many-ancestors
    """Dynamic dictionary to gather the required generic data from Netbox."""

    def _get_vlans(self) -> List[Dict[str, Any]]:
        """Returns all the vlans defined in Netbox.

        Returns:
            list: a list of vlans.

        """
        return [dict(i) for i in self._api.ipam.vlans.all()]


class NetboxDeviceData(BaseNetboxDeviceData):  # pylint: disable=too-many-ancestors
    """Dynamic dictionary to gather the required device-specific data from Netbox."""

    def _get_virtual_chassis_members(self) -> Optional[List[Dict[str, Any]]]:
        """Returns a list of devices part of the same virtual chassis or None.

        Returns:
            list: a list of devices.
            None: the device is not part of a virtual chassis.

        """
        if not self._device.metadata['netbox_object'].virtual_chassis:
            return None

        vc_id = self._device.metadata['netbox_object'].virtual_chassis.id
        return [dict(i) for i in self._api.dcim.devices.filter(virtual_chassis_id=vc_id)]

    def _get_circuits(self) -> Optional[Dict[str, Dict[str, Any]]]:
        """Returns a list of circuits connected to the device.

        Returns:
            list: A list of circuits.

        """
        # Because of changes documented in https://github.com/netbox-community/netbox/issues/4812
        # if an interface is connected to another device using a circuit, the circuit doesn't show up
        device_id = self._device.metadata['netbox_object'].id
        # Only get the circuits terminating where the device is
        circuits = {}

        # We get all the cables connected to a device
        for cable in self._api.dcim.cables.filter(device_id=device_id):
            # And if one side is a circuit we store it, with the local interface name as key
            if cable.termination_a_type == 'circuits.circuittermination':
                circuits[cable.termination_b.name] = dict(self._api.circuits.circuits.get(
                    cable.termination_a.circuit.id))
            elif cable.termination_b_type == 'circuits.circuittermination':
                circuits[cable.termination_a.name] = dict(self._api.circuits.circuits.get(
                    cable.termination_b.circuit.id))

        return circuits

    def _get_inventory(self) -> Optional[List[Dict[Any, Any]]]:
        """Returns the list of inventory items on the device.

        Returns:
            list: A list of inventory items.

        """
        device_id = self._device.metadata['netbox_object'].id
        return [dict(i) for i in self._api.dcim.inventory_items.filter(device_id=device_id)]

    def _get_vlans(self) -> Dict[int, Any]:
        """Returns all the vlans defined on a device.

        Returns:
            dict: a dict of vlans.

        """
        vlans = {}
        device_id = self._device.metadata['netbox_object'].id

        for interface in self._api.dcim.interfaces.filter(device_id=device_id):
            if interface.untagged_vlan and interface.untagged_vlan.vid not in vlans:
                vlans[interface.untagged_vlan.vid] = interface.untagged_vlan
            if interface.tagged_vlans:
                for tagged_vlan in interface.tagged_vlans:
                    if tagged_vlan.vid not in vlans:
                        vlans[tagged_vlan.vid] = tagged_vlan
        return vlans


class NetboxInventory:
    """Use Netbox as inventory to gather the list of devices to manage."""

    def __init__(self, api: pynetbox.api, device_roles: Sequence[str], device_statuses: Sequence[str]):
        """Initialize the instance.

        Arguments:
            api (pynetbox.api): the Netbox API instance.
            device_roles (list): a sequence of Netbox device role slug strings to filter the devices.
            device_statuses (list): a sequence of Netbox device status label or value strings to filter the devices.

        """
        self._api = api
        self._device_roles = device_roles
        self._device_statuses = [status.lower() for status in device_statuses]

    def get_devices(self) -> Dict[str, Dict[str, str]]:
        """Return all the devices based on configuration with their role and site.

        Returns:
            dict: a dictionary with the device FQDN as keys and a metadata dictionary as value.

        """
        devices = self._get_devices()
        devices.update(self._get_virtual_chassis_devices())
        return devices

    def _get_virtual_chassis_devices(self) -> Dict[str, Dict[str, str]]:
        """Get the devices part of virtual chassis according to the configuration.

        Returns:
            dict: a dictionary with the device FQDN as keys and a metadata dictionary as value.

        """
        devices: Dict[str, Dict[str, str]] = {}
        for vc in self._api.dcim.virtual_chassis.all():
            device = vc.master
            if not vc.domain:
                logger.error(
                    'Unable to determine hostname for virtual chassis of %s, domain property not set, skipping.',
                    device.name
                )
                continue
            if device.status.value not in self._device_statuses:
                logger.debug('Skipping device %s with status %s', device.name, device.status.label)
                continue
            if device.device_role.slug not in self._device_roles:
                logger.debug('Skipping device %s with role %s', device.name, device.device_role.slug)
                continue

            devices[vc.domain] = NetboxInventory._get_device_data(device)

        return devices

    def _get_devices(self) -> Dict[str, Dict[str, str]]:
        """Get the devices based on role, status and virtual chassis membership.

        Returns:
            dict: a dictionary with the device FQDN as keys and a metadata dictionary as value.

        """
        devices: Dict[str, Dict[str, str]] = {}
        for device in self._api.dcim.devices.filter(
                role=self._device_roles, status=self._device_statuses, virtual_chassis_member=False):
            if device.primary_ip4 is not None and device.primary_ip4.dns_name:
                fqdn = device.primary_ip4.dns_name
            elif device.primary_ip6 is not None and device.primary_ip6.dns_name:
                fqdn = device.primary_ip6.dns_name
            elif device.platform is None:
                logger.debug('Device %s missing FQDN and Platform, assuming non-manageable, skipping.', device.name)
                continue
            else:
                logger.error('Unable to determine FQDN for device %s, skipping.', device.name)
                continue

            devices[fqdn] = NetboxInventory._get_device_data(device)

        return devices

    @staticmethod
    def _get_device_data(device: pynetbox.models.dcim.Devices) -> Dict[str, str]:
        """Return the metadata needed from a Netbox device instance.

        Arguments:
            device (pynetbox.models.dcim.Devices): the Netbox device instance.

        Returns:
            dict: the dictionary with the device metadata.

        """
        metadata = {
            'role': device.device_role.slug,
            'site': device.site.slug,
            'type': device.device_type.slug,
            # Inject the Netbox object too to be future-proof and allow to get additional metadata without
            # the need of modifying homer's code. It also allow to use it inside NetboxData.
            'netbox_object': device,
        }

        # Convert Netbox interfaces into IPs
        if device.primary_ip4 is not None:
            metadata['ip4'] = ipaddress.ip_interface(device.primary_ip4.address).ip.compressed
        if device.primary_ip6 is not None:
            metadata['ip6'] = ipaddress.ip_interface(device.primary_ip6.address).ip.compressed

        return metadata
